"""Custom exceptions."""


class RuleExecutionError(Exception):
    """Rule fails during its execution."""

    pass


class ConditionExecutionError(Exception):
    """Condition fails during its execution."""

    pass



"""Module containing the class implementation of a rule.

Class: Rule
"""

from __future__ import annotations

import inspect
import logging
import re
from typing import Any, Callable
from warnings import warn

from arta.condition import BaseCondition, StandardCondition
from arta.exceptions import ConditionExecutionError, RuleExecutionError
from arta.utils import (
    ParsingErrorStrategy,
    parse_dynamic_parameter,
)

logger: logging.Logger = logging.getLogger(__name__)


class Rule:
    """A rule is the combination of some conditions and one action.

    Attributes:
        set_id: An id of a rule set.
        group_id: An id of a rule group.
        rule_id: The id of the rule.
        condition_exprs: A dictionary of condition expressions (k: condition conf. key, v: condition expression).
        condition_factory_mapping: A dictionary mapping between condition conf. keys and condition class objects
                                    (k: condition conf. key, v: condition class object).
        action: Function to perform when the conditions are valid (action function).
        action_parameters: Parameters of the action function.
    """

    def __init__(
        self,
        set_id: str,
        group_id: str,
        rule_id: str,
        condition_exprs: dict[str, str | None],
        condition_factory_mapping: dict[str, type[BaseCondition]],
        action: Callable,
        std_condition_instances: dict[str, StandardCondition],
        action_parameters: dict[str, Any] | None = None,
    ) -> None:
        """Initialize attributes.

        Args:
            std_condition_instances: Dictionary containing the BaseCondition instances required by the Rule
                (k: condition_id, v: StandardCondition instance).
        """
        # IDs
        self._set_id = set_id
        self._group_id = group_id
        self._rule_id = rule_id

        # Action
        self._action = action
        self._action_parameters = action_parameters or {}

        # Condition expressions
        self._condition_exprs = condition_exprs

        # Factory mapping
        self._condition_factory_mapping = condition_factory_mapping

        # Condition instances (k: condition id (not conf key), v: instances)
        self._condition_instances: dict[str, BaseCondition] = self._instantiate_conditions(std_condition_instances)

    def apply(
        self,
        input_data: dict[str, Any],
        *,
        parsing_error_strategy: ParsingErrorStrategy,
        **kwargs: Any,
    ) -> tuple[Any | None, dict[str, Any]]:
        """Apply the rule on the input data, return action output (optional).

        Args:
            input_data: Request or input data to apply rules on.
            parsing_error_strategy: Parsing error strategy.
            **kwargs: For user extra arguments.

        Returns:
            A tuple as: (action result, rule result details).

        Raises:
            RuleExecutionError: Error during the rule execution.
        """
        # If rule conditions are verified, the action is executed w/ the parameters' value
        is_conditions_ok: bool
        rule_results: dict[str, Any]

        is_conditions_ok, rule_results = self._check_conditions(
            input_data, parsing_error_strategy=parsing_error_strategy, **kwargs
        )

        if is_conditions_ok:
            logger.debug("Conditions are verified.")
            try:
                # Parse dynamic parameters
                parameters: dict[str, Any] = {}
                for key, value in self._action_parameters.items():
                    parameters[key] = parse_dynamic_parameter(
                        parameter=value, input_data=input_data, parsing_error_strategy=parsing_error_strategy
                    )

                # Track the rule id
                rule_results["activated_rule"] = self._rule_id

                # Pass input_data for value sharing if action function can accept it
                arg_spec: inspect.FullArgSpec = inspect.getfullargspec(self._action)
                if arg_spec.varkw is not None:
                    parameters["input_data"] = input_data
                    parameters.update(kwargs)

                # Backward compatibility case (now deprecated)
                if "input_data" in arg_spec.args or "input_data" in arg_spec.kwonlyargs:
                    warn(
                        (
                            "Using 'input_data' directly as an action function parameter is deprecated. "
                            "Use '**kwargs' instead. See how "
                            "at https://maif.github.io/arta/value_sharing/#between-conditions-and-actions"
                        ),
                        DeprecationWarning,
                        stacklevel=2,
                    )
                    parameters["input_data"] = input_data
                    parameters.update(kwargs)

                logger.debug(f"Action '{self._action.__name__}' is triggered.")

                # Run action
                rule_results["action_result"] = self._action(**parameters)

                return rule_results["action_result"], rule_results
            except Exception as error:
                msg: str = f"Error while executing rule '{self._rule_id}': {str(error)}"
                logger.error(msg)
                raise RuleExecutionError(msg) from error

        else:
            logger.debug("Conditions are not verified.")
            return None, {}

    def _check_conditions(
        self, input_data: dict[str, Any], parsing_error_strategy: ParsingErrorStrategy, **kwargs: Any
    ) -> tuple[bool, dict[str, Any]]:
        """(Protected)
        Return True if all conditions are verified.

        Args:
            input_data: Request or input data to apply rules on.
            parsing_error_strategy: Error handling strategy for parameter's parsing.
            **kwargs: For user extra arguments.

        Returns:
            A tuple as: (True if all conditions are verified, otherwise False, condition results dictionary).
        """
        # Var init.
        all_conditions_res: bool = True
        condition_results: dict[str, Any] = {"rule_group": self._group_id, "verified_conditions": {}}

        # Loop among condition expressions
        for cond_conf_key, expr in self._condition_exprs.items():
            condition_class: type[BaseCondition] = self._condition_factory_mapping[cond_conf_key]

            logger.debug(f"Verifying '{cond_conf_key}': {expr}")

            # Evaluate the condition expression
            try:
                condition_res, unitary_res = self._evaluate_condition_expr(
                    input_data=input_data,
                    condition_class=condition_class,
                    condition_expr=expr,
                    parsing_error_strategy=parsing_error_strategy,
                    **kwargs,
                )
            except NameError as e:
                msg: str = f"Error during evaluation of '{cond_conf_key}: {expr}': {str(e)}"
                logger.error(msg)
                raise RuleExecutionError(msg) from e

            # Combine conditions (AND)
            all_conditions_res = all_conditions_res and condition_res

            # Store condition results
            condition_results["verified_conditions"].update(
                {cond_conf_key: {"expression": expr, "values": unitary_res}}
            )

            if not all_conditions_res:
                # If False, no need to go further
                break

        return all_conditions_res, condition_results

    def _evaluate_condition_expr(
        self,
        input_data: dict[str, Any],
        condition_class: type[BaseCondition],
        parsing_error_strategy: ParsingErrorStrategy,
        condition_expr: str | None = None,
        **kwargs: Any,
    ) -> tuple[bool, dict[str, bool]]:
        """(Protected)
        Evaluate the condition expr (a boolean expression) and
        return the result (a boolean).

        Args:
            input_data: Request or input data.
            condition_class: Class object of the analyzed condition (given by its conf. key).
            parsing_error_strategy: Error handling strategy for parameter's parsing.
            condition_expr: A boolean expression (string).
            **kwargs: For user extra arguments.

        Returns:
            A tuple as: (final result, unitary results (dictionary)).

        Raises:
            ConditionExecutionError: Error during condition execution.
        """
        # Var init.
        unitary_results: dict[str, bool] = {}

        # Case of null condition expressions => Always True
        if condition_expr is None:
            return True, unitary_results

        # The final boolean expression is formed by replacing condition IDs by their evaluated boolean values
        bool_expr: str = condition_expr

        # Extract condition ids from the expression
        condition_ids: set[str] = condition_class.extract_condition_ids_from_expression(condition_expr)

        # Loop among the conditions of the expression
        # Verify the unitary condition
        for cond_id in condition_ids:
            # Retrieve condition instance
            condition: BaseCondition = self._condition_instances[cond_id]

            # Check unitary condition
            try:
                bool_var: bool = condition.verify(input_data, parsing_error_strategy=parsing_error_strategy, **kwargs)

                # Store unitary result
                unitary_results[cond_id] = bool_var
            except Exception as error:
                msg: str = f"Error while executing condition '{cond_id}': {str(error)}"
                logger.error(msg)
                raise ConditionExecutionError(msg) from error

            # Replace the result in the boolean expression
            sanit_cond_id: str = condition.get_sanitized_id()
            bool_expr = re.sub(rf"{sanit_cond_id}", str(bool_var), bool_expr)

        # Evaluate the final boolean expressions = final result
        return eval(bool_expr), unitary_results  # noqa

    def _instantiate_conditions(
        self,
        std_conditions: dict[str, StandardCondition],
    ) -> dict[str, BaseCondition]:
        """Parse condition expressions and build corresponding instances.

        E.g., for one condition:
        - Input : "not(CONDITION_A) and CONDITION_B"
        - Output : {"CONDITION_A": condition_A_instance, "CONDITION_B": condition_B_instance}

        Args:
            std_conditions: A dictionary containing the StandardCondition instances
                (k: cond. id, v: StandardCondition instance)

        Returns:
            Condition instances which are in the condition expressions (k: condition id, v: BaseCondition instance).

        Raises:
            KeyError: Condition id is unknown.
        """
        # Var init.
        condition_ids: set[str] = set()
        cond_instances: dict[str, BaseCondition] = {}

        # Get all condition ids from the expressions (1 or many)
        for conf_key, expr in self._condition_exprs.items():
            # Expression parsing is condition class dependent
            condition_ids = self._condition_factory_mapping[conf_key].extract_condition_ids_from_expression(expr)

            # Is a custom condition or a simple condition?
            if self._condition_factory_mapping is not None and conf_key != "condition":
                # Yes
                for cond_id in condition_ids:
                    # Instanciate the custom (unknown) condition object
                    cond_instances[cond_id] = self._condition_factory_mapping[conf_key](
                        condition_id=cond_id,
                        description=self._condition_factory_mapping[conf_key].CONDITION_DATA_LABEL,
                    )
            else:
                # Should be a standard condition
                for cond_id in condition_ids:
                    # Retrieve the standard condition instance (already instantiated)
                    try:
                        cond_instances[cond_id] = std_conditions[cond_id]
                    except KeyError as error:
                        msg: str = f"Following condition id is unknown '{cond_id}' in {conf_key}: {expr}"
                        logger.error(msg)
                        raise KeyError(msg) from error

        return cond_instances



"""Module implementing the rules engine.

Class: RulesEngine
"""

from __future__ import annotations

import copy
import importlib
import inspect
import logging
from inspect import getmembers, isclass, isfunction
from pathlib import Path
from types import FunctionType, MethodType, ModuleType
from typing import Any, Callable

from arta.condition import BaseCondition, SimpleCondition, StandardCondition
from arta.config import load_config
from arta.models import Configuration, RulesDict
from arta.rule import Rule
from arta.utils import ParsingErrorStrategy, RuleActivationMode

logger: logging.Logger = logging.getLogger(__name__)


class RulesEngine:
    """The Rules Engine is in charge of executing different groups of rules of a given rule set on user input data.

    Attributes:
        rules:  A dictionary of rules with k: rule set, v: (k: rule group, v: list of rule instances).
    """

    # ==== Class constants ====

    # Rule related config keys
    CONST_RULE_SETS_CONF_KEY: str = "rules"
    CONST_DFLT_RULE_SET_ID: str = "default_rule_set"
    CONST_STD_RULE_CONDITION_CONF_KEY: str = "condition"
    CONST_ACTION_CONF_KEY: str = "action"
    CONST_ACTION_PARAMETERS_CONF_KEY: str = "action_parameters"

    # Condition related config keys
    CONST_STD_CONDITIONS_CONF_KEY: str = "conditions"
    CONST_CONDITION_VALIDATION_FUNCTION_CONF_KEY: str = "validation_function"
    CONST_CONDITION_DESCRIPTION_CONF_KEY: str = "description"
    CONST_CONDITION_VALIDATION_PARAMETERS_CONF_KEY: str = "condition_parameters"
    CONST_USER_CONDITION_STRING: str = "USER_CONDITION"

    # Built-in factory mapping
    BUILTIN_FACTORY_MAPPING: dict[str, type[BaseCondition]] = {
        "condition": StandardCondition,
        "simple_condition": SimpleCondition,
    }

    def __init__(
        self,
        *,
        rules_dict: dict[str, dict[str, Any]] | None = None,
        config_path: Path | str | None = None,
        config_dict: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the rules.

        3 possibilities: either one of 'rules_dict', or 'config_path', or 'config_dict'.

        Args:
            rules_dict: A dictionary containing the rules' definitions.
            config_path: Path to the directory containing the YAML files.
            config_dict: A dictionary containing the configuration (same as YAML files but already
                         parsed in a dictionary).

        Raises:
            KeyError: Key not found.
            TypeError: Wrong type.
            ValueError: Bad given parameters.
        """
        # Var init.
        factory_mapping_classes: dict[str, type[BaseCondition]] = {}
        std_condition_instances: dict[str, StandardCondition] = {}

        given_params: list[bool] = [config_path is not None, rules_dict is not None, config_dict is not None]

        if given_params.count(True) != 1:
            msg: str = "RulesEngine takes one (and only one) parameter: 'rules_dict' or 'config_path' or 'config_dict'."
            logger.error(msg)
            raise ValueError(msg)

        # Init. default global settings (useful if not set, can't be set in the Pydantic model
        # because of the rules dict mode)
        self._parsing_error_strategy: ParsingErrorStrategy = ParsingErrorStrategy.RAISE
        self._rule_activation_mode: RuleActivationMode = RuleActivationMode.ONE_BY_GROUP

        # Initialize directly with a rules dict
        if rules_dict is not None:
            # Data validation
            # RulesDict.parse_obj(rules_dict)
            RulesDict.model_validate(rules_dict)

            # Edge cases data validation
            if not isinstance(rules_dict, dict):
                msg = f"'rules_dict' must be dict type, not '{type(rules_dict)}'."
                logger.error(msg)
                raise TypeError(msg)
            elif len(rules_dict) == 0:
                msg = "'rules_dict' couldn't be empty."
                logger.error(msg)
                raise KeyError(msg)

            # Attribute definition
            self.rules: dict[str, dict[str, list[Rule]]] = self._adapt_user_rules_dict(rules_dict)

        # Initialize with a config_path or config_dict
        else:
            if config_path is not None:
                # Load config in attribute
                config_dict = load_config(config_path)

            # Data validation
            config: Configuration = Configuration.model_validate(config_dict)

            if config.parsing_error_strategy is not None:
                # Set parsing error handling strategy from config
                self._parsing_error_strategy = ParsingErrorStrategy(config.parsing_error_strategy)

            if config.rule_activation_mode is not None:
                # Set rule activation mode from config
                self._rule_activation_mode = RuleActivationMode(config.rule_activation_mode)

            # dict of available action functions (k: function name, v: function object)
            action_modules: list[str] = config.actions_source_modules
            action_functions: dict[str, Callable] = self._get_object_from_source_modules(action_modules)

            # dict of available standard condition functions (k: function name, v: function object)
            condition_modules: list[str] = (
                config.conditions_source_modules if config.conditions_source_modules is not None else []
            )
            std_condition_functions: dict[str, Callable] = self._get_object_from_source_modules(condition_modules)

            # Dictionary of condition instances (k: condition id, v: instance), built from config data
            if len(std_condition_functions) > 0:
                std_condition_instances = self._build_std_conditions(
                    config=config.model_dump(), condition_functions_dict=std_condition_functions
                )

            # User-defined/custom conditions
            if config.condition_factory_mapping is not None and config.custom_classes_source_modules is not None:
                logger.info("Custom condition configuration detected.")

                # dict of custom condition classes (k: classe name, v: class object)
                custom_condition_classes: dict[str, type[BaseCondition]] = self._get_object_from_source_modules(
                    config.custom_classes_source_modules
                )

                # Build a factory mapping dictionary (k: conf key, v:class object)
                factory_mapping_classes.update(
                    {
                        conf_key: custom_condition_classes[class_name]
                        for conf_key, class_name in config.condition_factory_mapping.items()
                    }
                )

            # Arta built-in conditions
            factory_mapping_classes.update(self.BUILTIN_FACTORY_MAPPING)

            # Attribute definition
            self.rules = self._build_rules(
                std_condition_instances=std_condition_instances,
                action_functions=action_functions,
                config=config.model_dump(),
                factory_mapping_classes=factory_mapping_classes,
            )

        logger.info(
            f"Rules engine correctly instanciated with '{str(self._parsing_error_strategy)}' and '{str(self._rule_activation_mode)}'"
        )

    def apply_rules(
        self,
        input_data: dict[str, Any],
        *,
        rule_set: str | None = None,
        ignored_rules: set[str] | None = None,
        verbose: bool = False,
        **kwargs: Any,
    ) -> dict[str, Any]:
        """Apply the rules and return results.

        For each rule group of a given rule set, rules are applied sequentially,
        The loop is broken when a rule is applied (an action is triggered)
        or not (depends on the rule activation mode).
        Then, the next rule group is evaluated.
        And so on...

        This means that the order of the rules in the configuration file
        (e.g., rules.yaml) is meaningful.

        Args:
            input_data: Input data to apply rules on.
            rule_set: Apply rules associated with the specified rule set.
            ignored_rules: A set/list of rule's ids to be ignored/disabled during evaluation.
            verbose: If True, add extra ids (group_id, rule_id) for result explicability.
            **kwargs: For user extra arguments.

        Returns:
            A dictionary containing the rule groups' results (k: group id, v: action result).

        Raises:
            TypeError: Wrong type (e.g., input_data is not a dictionary).
            KeyError: Key not found (e.g., input_data is an empty dictionary).
            RuleExecutionError: A rule fails during execution.
            ConditionExecutionError: A condition fails during execution.
        """
        rule_count: int = 0

        # Input_data validation
        if not isinstance(input_data, dict):
            msg: str = f"'input_data' must be dict type, not '{type(input_data)}'."
            logger.error(msg)
            raise TypeError(msg)
        elif len(input_data) == 0:
            msg = "'input_data' couldn't be empty."
            logger.error(msg)
            raise KeyError(msg)

        # Var init.
        input_data_copy: dict[str, Any] = copy.deepcopy(input_data)
        ignored_ids: set[str] = ignored_rules if ignored_rules is not None else set()
        if len(ignored_ids) > 0:
            logger.info(f"Configured ignored rules are: {ignored_ids}")

        # Prepare the result key
        input_data_copy["output"] = {}

        # If there is no given rule set param. and there is only one rule set in self.rules
        # and its value is 'default_rule_set', look for this one (rule_set='default_rule_set')
        if rule_set is None and len(self.rules) == 1 and self.rules.get(self.CONST_DFLT_RULE_SET_ID) is not None:
            rule_set = self.CONST_DFLT_RULE_SET_ID

        logger.info(f"Rules engine is running with the following rule set: '{rule_set}', verbose: {verbose}")

        # Check if given rule set is in self.rules?
        if rule_set not in self.rules:
            msg = f"Rule set '{rule_set}' not found in the rules, available rule sets are : {list(self.rules.keys())}."
            logger.error(msg)
            raise KeyError(msg)

        # Var init.
        results_dict: dict[str, Any] = {"verbosity": {"rule_set": rule_set, "results": []}}

        # Groups' loop
        for group_id, rules_list in self.rules[rule_set].items():
            group_rule_count: int = 0
            logger.debug(f"Entering rule group: {group_id}")

            # Initialize result of the rule group with None
            results_dict[group_id] = None

            # Rules' loop (inside a group)
            for rule in rules_list:
                if rule._rule_id in ignored_ids:
                    # Ignore that rule
                    continue

                rule_count += 1
                group_rule_count += 1
                logger.debug(f"Evaluating rule '{group_rule_count}': {rule._rule_id}")

                # Apply rules
                action_result, rule_details = rule.apply(
                    input_data_copy, parsing_error_strategy=self._parsing_error_strategy, **kwargs
                )

                # Check if the rule has been applied (= action activated)
                if "action_result" in rule_details:
                    # Save result and details
                    results_dict[group_id] = action_result
                    results_dict["verbosity"]["results"].append(rule_details)

                    # Update input data with current result with key 'output' (can be used in next rules)
                    input_data_copy["output"][group_id] = copy.deepcopy(results_dict[group_id])

                    if self._rule_activation_mode is RuleActivationMode.ONE_BY_GROUP:
                        # We can only have one result per group => break when "action_result" in rule_details
                        break

        # Handling non-verbose mode
        if not verbose:
            results_dict.pop("verbosity")

        logger.info(f"'{rule_count}' rules were correctly evaluated against input data.")
        return results_dict

    @staticmethod
    def _get_object_from_source_modules(module_list: list[str]) -> dict[str, Any]:
        """(Protected)
        Collect all functions defined in the list of modules.

        Args:
            module_list: List of source module names.

        Returns:
            Dictionary with objects found in the modules.
        """
        object_dict: dict[str, Any] = {}

        for module_name in module_list:
            # Import module
            mod: ModuleType = importlib.import_module(module_name)

            # Collect functions
            module_functions: dict[str, Any] = {key: val for key, val in getmembers(mod, isfunction)}
            object_dict.update(module_functions)

            # Collect classes
            module_classes: dict[str, Any] = {key: val for key, val in getmembers(mod, isclass)}
            object_dict.update(module_classes)

        return object_dict

    def _build_rules(
        self,
        std_condition_instances: dict[str, StandardCondition],
        action_functions: dict[str, Callable],
        config: dict[str, Any],
        factory_mapping_classes: dict[str, type[BaseCondition]],
    ) -> dict[str, dict[str, list[Any]]]:
        """(Protected)
        Return a dictionary of Rule instances built from the configuration.

        Args:
            std_condition_instances: Dictionary of condition instances (k: condition id, v: StandardCondition instance)
            action_functions: Dictionary of action functions (k: action name, v: Callable)
            config: Dictionary of the imported configuration from yaml files.
            factory_mapping_classes: A mapping dictionary (k: condition conf. key, v: custom class object)

        Returns:
            A dictionary of rules.
        """
        # Var init.
        rules_dict: dict[str, dict[str, list[Any]]] = {}

        # Retrieve rule set ids from config
        rule_set_ids: list[str] = list(config[self.CONST_RULE_SETS_CONF_KEY].keys())

        # Going all way down to the rules (rule set > rule group > rule)
        for set_id in rule_set_ids:
            rules_conf: dict[str, Any] = config[self.CONST_RULE_SETS_CONF_KEY][set_id]
            rules_dict[set_id] = {}
            rule_set_dict: dict[str, list[Any]] = rules_dict[set_id]

            # Looping throught groups
            for group_id, group_rules in rules_conf.items():
                # Initialize list or rules in the group
                rule_set_dict[group_id] = []

                # Looping through rules (inside a group)
                for rule_id, rule_dict in group_rules.items():
                    # Get action function
                    action_function_name: str = rule_dict[self.CONST_ACTION_CONF_KEY]

                    if action_function_name not in action_functions:
                        msg: str = f"Unknwown action function : {action_function_name}"
                        logger.error(msg)
                        raise KeyError(msg)

                    action: Callable = action_functions[action_function_name]

                    # Look for condition conf. keys inside the rule
                    condition_conf_keys: set[str] = set(rule_dict.keys()) - {
                        self.CONST_ACTION_CONF_KEY,
                        self.CONST_ACTION_PARAMETERS_CONF_KEY,
                    }

                    # Store the cond. expressions with the same order as in the configuration file (very important)
                    condition_exprs: dict[str, str | None] = {
                        key: value for key, value in rule_dict.items() if key in condition_conf_keys
                    }

                    # Create the corresponding Rule instance
                    rule: Rule = Rule(
                        set_id=set_id,
                        group_id=group_id,
                        rule_id=rule_id,
                        action=action,
                        action_parameters=rule_dict[self.CONST_ACTION_PARAMETERS_CONF_KEY],
                        condition_exprs=condition_exprs,
                        std_condition_instances=std_condition_instances,
                        condition_factory_mapping=factory_mapping_classes,
                    )
                    rule_set_dict[group_id].append(rule)

        return rules_dict

    def _build_std_conditions(
        self, config: dict[str, Any], condition_functions_dict: dict[str, Callable]
    ) -> dict[str, StandardCondition]:
        """(Protected)
        Return a dictionary of Condition instances built from the configuration file.

        Args:
            config: Dictionary of the imported configuration from yaml files.
            condition_functions_dict: A dictionary where k:condition id, v:Callable (validation function).

        Returns:
            A dictionary of StandardCondition instances (k: condition id, v: StandardCondition instance).
        """
        # Var init.
        conditions_dict: dict[str, StandardCondition] = {}

        # Condition configuration (under conditions' key)
        conditions_conf: dict[str, dict[str, Any]] = config[self.CONST_STD_CONDITIONS_CONF_KEY]

        # Looping through conditions (inside a group)
        for condition_id, condition_params in conditions_conf.items():
            # Get condition validation function
            validation_function_name: str = condition_params[self.CONST_CONDITION_VALIDATION_FUNCTION_CONF_KEY]

            if validation_function_name not in condition_functions_dict:
                msg: str = f"Unknwown validation function : {validation_function_name}"
                logger.error(msg)
                raise KeyError(msg)

            # Get Callable from function name
            validation_function: Callable = condition_functions_dict[validation_function_name]

            # Create Condition instance
            condition_instance: StandardCondition = StandardCondition(
                condition_id=condition_id,
                description=condition_params[self.CONST_CONDITION_DESCRIPTION_CONF_KEY],
                validation_function=validation_function,
                validation_function_parameters=condition_params[self.CONST_CONDITION_VALIDATION_PARAMETERS_CONF_KEY],
            )
            conditions_dict[condition_id] = condition_instance

        return conditions_dict

    def _adapt_user_rules_dict(self, rules_dict: dict[str, dict[str, Any]]) -> dict[str, dict[str, list[Any]]]:
        """(Protected)
        Return a dictionary of Rule's instances built from user's rules dictionary.

        Args:
            rules_dict: User raw rules dictionary.

        Returns:
            A rules dictionary made from the user input rules.
        """
        # Var init.
        rules_dict_formatted: dict[str, list[Any]] = {}

        # Looping throught groups
        for group_id, group_rules in rules_dict.items():
            # Initialize list or rules in the group
            rules_dict_formatted[group_id] = []

            # Looping through rules (inside a group)
            for rule_id, rule_dict in group_rules.items():
                # Get action function
                action = rule_dict["action"]

                # Create Rule instance
                rule = Rule(
                    set_id=self.CONST_DFLT_RULE_SET_ID,
                    group_id=group_id,
                    rule_id=rule_id,
                    action=action,
                    action_parameters=rule_dict.get(self.CONST_ACTION_PARAMETERS_CONF_KEY),
                    condition_exprs={self.CONST_STD_RULE_CONDITION_CONF_KEY: self.CONST_USER_CONDITION_STRING}
                    if self.CONST_STD_RULE_CONDITION_CONF_KEY in rule_dict
                    and rule_dict.get(self.CONST_STD_RULE_CONDITION_CONF_KEY) is not None
                    else {self.CONST_STD_RULE_CONDITION_CONF_KEY: None},
                    std_condition_instances={
                        self.CONST_USER_CONDITION_STRING: StandardCondition(
                            condition_id=self.CONST_USER_CONDITION_STRING,
                            description="Automatic description",
                            validation_function=rule_dict.get(self.CONST_STD_RULE_CONDITION_CONF_KEY),
                            validation_function_parameters=rule_dict.get(
                                self.CONST_CONDITION_VALIDATION_PARAMETERS_CONF_KEY
                            ),
                        )
                    },
                    condition_factory_mapping=self.BUILTIN_FACTORY_MAPPING,
                )
                rules_dict_formatted[group_id].append(rule)

        return {self.CONST_DFLT_RULE_SET_ID: rules_dict_formatted}

    def __str__(self) -> str:
        """Object human string representation (called by str()).

        Returns:
            A string representation of the instance.
        """
        # Vars init.
        attrs_str: str = ""

        # Get some instance attributes infos
        class_name: str = self.__class__.__name__
        attrs: list[tuple[str, Any]] = [
            attr
            for attr in inspect.getmembers(self)
            if not (
                attr[0].startswith("_")
                or attr[0].startswith("CONST_")
                or isinstance(attr[1], (FunctionType, MethodType))
            )
        ]

        # Build string representation
        for attr, val in attrs:
            attrs_str += f"{attr}={str(val)}, "

        return f"{class_name}({attrs_str})"



"""Pydantic model implementations."""

from __future__ import annotations

from typing import Annotated, Any, Callable, Optional
from warnings import warn

import pydantic
from pydantic.version import VERSION

from arta.utils import ParsingErrorStrategy, RuleActivationMode

PYDANTIC_V1: bool = VERSION.startswith("1.")

if not PYDANTIC_V1:
    # Pydantic V2

    # ----------------------------------
    # For instantiation using rules_dict
    class RuleRaw(pydantic.BaseModel):
        """Pydantic model for validating a rule."""

        condition: Optional[Callable] = None
        condition_parameters: Optional[dict[str, Any]] = None
        action: Callable
        action_parameters: Optional[dict[str, Any]] = None

        model_config = pydantic.ConfigDict(extra="forbid")

    class RulesGroup(pydantic.RootModel):
        """Pydantic model for validating a rules group."""

        root: dict[str, RuleRaw]

    class RulesDict(pydantic.RootModel):
        """Pydantic model for validating rules dict instanciation."""

        root: dict[str, RulesGroup]

    # ----------------------------------
    # For instantiation using config_path
    class Condition(pydantic.BaseModel):
        """Pydantic model for validating a condition."""

        description: str
        validation_function: str
        condition_parameters: Optional[dict[str, Any]] = None

    class RulesConfig(pydantic.BaseModel):
        """Pydantic model for validating a rule group from config file."""

        condition: Optional[str] = None
        simple_condition: Optional[str] = None
        action: Annotated[str, pydantic.StringConstraints(to_lower=True)]
        action_parameters: Optional[Any] = None

        model_config = pydantic.ConfigDict(extra="allow")

    class Configuration(pydantic.BaseModel):
        """Pydantic model for validating configuration files."""

        conditions: Optional[dict[str, Condition]] = None
        conditions_source_modules: Optional[list[str]] = None
        actions_source_modules: list[str]
        custom_classes_source_modules: Optional[list[str]] = None
        condition_factory_mapping: Optional[dict[str, str]] = None
        rules: dict[str, dict[str, dict[Annotated[str, pydantic.StringConstraints(to_upper=True)], RulesConfig]]]
        parsing_error_strategy: Optional[ParsingErrorStrategy] = None
        rule_activation_mode: Optional[RuleActivationMode] = None

else:
    # Pydantic V1

    warn(
        (
            "Soon, Pydantic V1 will no longer be compatible with Arta. "
            "Please, migrate to Pydantic V2 (https://docs.pydantic.dev/latest/migration/)."
        ),
        DeprecationWarning,
        stacklevel=2,
    )

    class BaseModelV2(pydantic.BaseModel):
        """Wrapper to expose missed methods used elsewhere in the code"""

        model_dump: Callable = pydantic.BaseModel.dict

        @classmethod
        def model_validate(cls, obj):
            """Method mapping between V1 to V2."""
            return cls.parse_obj(obj)

    # ----------------------------------
    # For instantiation using rules_dict
    class RuleRaw(BaseModelV2):  # type: ignore[no-redef]
        """Pydantic model for validating a rule."""

        condition: Optional[Callable]
        condition_parameters: Optional[dict[str, Any]]
        action: Callable
        action_parameters: Optional[dict[str, Any]]

        class Config:
            extra = "forbid"

    class RulesGroup(pydantic.BaseModel):  # type: ignore[no-redef]
        """Pydantic model for validating a rules group."""

        __root__: dict[str, RuleRaw]

    class RulesDict(BaseModelV2):  # type: ignore[no-redef]
        """Pydantic model for validating rules dict instanciation."""

        __root__: dict[str, RulesGroup]

    # ----------------------------------
    # For instantiation using config_path
    class Condition(BaseModelV2):  # type: ignore[no-redef]
        """Pydantic model for validating a condition."""

        description: str
        validation_function: str
        condition_parameters: Optional[dict[str, Any]]

    class RulesConfig(BaseModelV2):  # type: ignore[no-redef]
        """Pydantic model for validating a rule group from config file."""

        condition: Optional[str]
        simple_condition: Optional[str]
        action: pydantic.constr(to_lower=True)  # type: ignore
        action_parameters: Optional[Any]

        class Config:
            extra = "allow"

    class Configuration(BaseModelV2):  # type: ignore[no-redef]
        """Pydantic model for validating configuration files."""

        conditions: Optional[dict[str, Condition]]
        conditions_source_modules: Optional[list[str]]
        actions_source_modules: list[str]
        custom_classes_source_modules: Optional[list[str]]
        condition_factory_mapping: Optional[dict[str, str]]
        rules: dict[str, dict[str, dict[pydantic.constr(to_upper=True), RulesConfig]]]  # type: ignore
        parsing_error_strategy: Optional[ParsingErrorStrategy] = None
        rule_activation_mode: Optional[RuleActivationMode] = None
